"""Module to detect slow waves.

"""
from logging import getLogger
from numpy import argmax, concatenate, hstack, sum, zeros, median, mean

from .spindle import detect_events, transform_signal, within_duration
from ..graphoelement import SlowWaves

lg = getLogger(__name__)
MAXIMUM_DURATION = 5

class DetectSlowWave:
    """Design slow wave detection on a single channel.
    
    Parameters
    ----------
    method : str
        one of the predefined methods
    frequency : tuple of float
        low and high frequency of SW band
    duration : tuple of float
        min and max duration of SWs        

    """
    def __init__(self, method='Massimini2004', duration=None):
        
        if duration is None:
            duration = (0.5, 3)
            
        self.method = method
        self.duration = duration
        
        if method == 'Massimini2004':
            self.det_butter = {'order': 4,
                               'freq': (0.1, 4.)}
            self.trough_duration = (0.3, 1.)
            self.max_trough_amp = - 80
            self.min_ptp = 140
            
        elif method == 'AASM/Massimini2004':
            self.det_butter = {'order': 4,
                               'freq': (0.1, 4.)}
            self.trough_duration = (0.25, 1.)
            self.max_trough_amp = - 40
            self.min_ptp = 75
        
        else:
            raise ValueError('Unknown method')
            
    def __repr__(self):
        return ('detsw_{0}_{3:04.1f}-{4:04.1f}s'
                ''.format(self.method, self.duration[0], self.duration[1]))
    
    def __call__(self, data):
        """Detect slow waves on the data.

        Parameters
        ----------
        data : instance of Data
            data used for detection

        Returns
        -------
        instance of graphoelement.SlowWaves
            description of the detected SWs

        """
        slowwave = SlowWaves()
        slowwave.chan_name = data.axis['chan'][0]

        all_slowwaves = []
        for i, chan in enumerate(data.axis['chan'][0]):
            lg.info('Detecting slow waves on chan %s', chan)
            time = hstack(data.axis['time'])
            dat_orig = hstack(data(chan=chan))
            
            if self.method == 'Massimini2004':
                sw_in_chan = detect_Massimini2004(
                        dat_orig, data.s_freq, time, self)
                
            elif self.method == 'AASM/Massimini2004':
                sw_in_chan = detect_Massimini2004(
                        dat_orig, data.s_freq, time, self)
                
            else:
                raise ValueError('Unknown method')

            for sw in sw_in_chan:
                sw.update({'chan': chan})
            all_slowwaves.extend(sw_in_chan)
            # end of loop over chan
            
        lg.info('number of SW: ' + str(len(all_slowwaves)))
        slowwave.events = sorted(all_slowwaves, key=lambda x: x['start'])
        
        return slowwave
    
def detect_Massimini2004(dat_orig, s_freq, time, opts):
    """Slow wave detection based on Massimini et al., 2004.
    
    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSlowWave'
        'det_butter' : dict
            parameters for 'butter',
        'duration' : tuple of float
            min and max duration of SW
        'min_ptp' : float
            min peak-to-peak amplitude   
        'trough_duration' : tuple of float
            min and max duration of first half-wave (trough)

    Returns
    -------
    list of dict
        list of detected SWs
    float
        SW density, per 30-s epoch
        
    References
    ----------
    Massimini, M. et al. J Neurosci 24(31) 6862-70 (2004).
    
    """
    lg.info('detection raw dat: ' + str(len(dat_orig)))
    lg.info('mean: ' + str(mean(dat_orig)) +' median: ' + str(median(dat_orig)))
    dat_det = transform_signal(dat_orig, s_freq, 'butter', opts.det_butter)
    lg.info('det filtered data: ' + str(len(dat_det)))
    lg.info('mean: ' + str(mean(dat_det)) +' median: ' + str(median(dat_det)))
    below_zero = detect_events(dat_det, 'upper_threshold', value=0.)
    
    if below_zero is not None:
        troughs = within_duration(below_zero, time, opts.trough_duration) 
        
        if troughs is not None:
            troughs = select_peaks(dat_det, troughs, opts.max_trough_amp)
            
            if troughs is not None:
                events = _add_pos_halfwave(dat_det, troughs, s_freq, opts)
                
                if len(events):
                    events = within_duration(events, time, opts.duration)
                    
                    sw_in_chan = make_slow_waves(events, dat_det, time, s_freq)

    else:
        lg.info('No slow wave found')
        sw_in_chan = []
        
    return sw_in_chan
    

def select_peaks(data, events, limit):
    """Check whether event is satisfies amplitude limit.

    Parameters
    ----------
    data : ndarray (dtype='float')
        vector with data
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak/trough, end samples
    limit : float
        low and high limit for spindle duration

    Returns
    -------
    ndarray (dtype='int')
        N x 2+ matrix with start, peak, end samples

    """
    selected = abs(data[events[:, 1]]) >= abs(limit)

    return events[selected, :]

def make_slow_waves(events, data, time, s_freq):
    """Create dict for each slow wave, based on events of time points.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x 5 matrix with start, trough, zero, peak, end samples
    data : ndarray (dtype='float')
        vector with the data
    time : ndarray (dtype='float')
        vector with time points
    s_freq : float
        sampling frequency

    Returns
    -------
    list of dict
        list of all the SWs, with information about start_time, 
        trough_time, zero_time, peak_time, end_time, duration (s), trough_val, 
        peak_val, peak-to-peak amplitude (signal units), area_under_curve 
        (signal units * s)
    """
    slow_waves = []
    for ev in events:
        one_sw = {'start': time[ev[0]],
                  'trough_time': time[ev[1]],
                  'zero_time': time[ev[2]],
                  'peak_time': time[ev[3]],
                  'end': time[ev[4]-1],
                  'trough_val': data[ev[1]],
                  'peak_val': data[ev[3]],
                  'dur': (ev[4] - ev[0]) / s_freq,
                  'area_under_curve': sum(data[ev[0]: ev[4]]) / s_freq,
                  'ptp': ev[3] - ev[1]
                  }
        slow_waves.append(one_sw)

    return slow_waves

def _add_pos_halfwave(data, events, s_freq, opts):
    """Find the next zero crossing and the intervening peak and add them to 
    events. If no zero found before max_dur, event is discarded. If 
    peak-to-peak is smaller than min_ptp, the event is discarded.
    
    Parameters
    ----------
    data : ndarray (dtype='float')
        vector with the data
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples    s_freq : float
        sampling frequency
    opts : instance of 'DetectSlowWave'
        'duration' : tuple of float
            min and max duration of SW
        'min_ptp' : float
            min peak-to-peak amplitude   
            
    Returns
    -------
    ndarray (dtype='int')
        N x 5 matrix with start, trough, - to + zero crossing, peak, and end 
        samples
    """
    max_dur = opts.duration[1]
    
    if max_dur is None:
        max_dur = MAXIMUM_DURATION    
    window = s_freq * max_dur
    
    peak_and_zero = zeros((events.shape[0], 2))
    events = concatenate((events, peak_and_zero), axis=1)
    selected = []
    
    for ev in events:
        ev[4] = ev[2] + argmax(data[ev[2]: ev[0] + window] < 0) # quickest way
        
        if ev[4] == 0:
            selected.append(False)
            continue
        
        ev[3] = argmax(data[ev[2]: ev[4]])
        
        if abs(data[ev[1]] - data[ev[3]]) < opts.min_p2p:
            selected.append(False)
            continue
        
        selected.append(True)
            
    return events[selected, :]
            
        
    

    


"""Package to import and export common formats.

"""
from .edf import Edf, write_edf
from .ktlx import Ktlx
from .blackrock import BlackRock
from .egimff import EgiMff
from .moberg import Moberg
from .mnefiff import write_mnefiff
from .fieldtrip import FieldTrip, write_fieldtrip
from .phypno import Phypno, write_phypno
from .ieeg_org import IEEG_org
from .opbox import OpBox
# from .eeglab import Eeglab, write_eeglab



"""Module to plot all the elements in 3d space.
"""
from numpy import array, isnan, max, mean, min, nanmax, nanmin, r_
from vispy.color import get_colormap
from vispy.geometry import create_sphere, MeshData
from vispy.visuals.transforms import STTransform

from .base import COLORMAP, normalize, SimpleMesh, Viz


CHAN_COLOR = 0., 1, 0, 1.
SKIN_COLOR = 0.94, 0.82, 0.81, 1.

SCALE_FACTOR = 135
ELEVATION = 0


class Viz3(Viz):
    """The 3d visualization, ordinarily it should hold a surface and electrodes

    There is only one plot, so we define it at the __init__
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._plt = self._fig[0, 0]
        self._plt._configure_3d()

        # remove white space around the main plot
        self._plt.grid.margin = 0
        self._plt.title.stretch = 0, 0

        self._surf = []
        self._chan = []

        self._chan_colormap = None
        self._chan_limits_c = None

    def add_surf(self, surf, color=SKIN_COLOR, vertex_colors=None,
                 values=None, limits_c=None, colormap=COLORMAP):
        """Add surfaces to the visualization.

        Parameters
        ----------
        surf : instance of sleepytimes.attr.anat.Surf
            surface to be plotted
        color : tuple or ndarray, optional
            4-element tuple, representing RGB and alpha, between 0 and 1
        vertex_colors : ndarray
            ndarray with n vertices x 4 to specify color of each vertex
        values : ndarray, optional
            vector with values for each vertex
        limits_c : tuple of 2 floats, optional
            min and max values to normalize the color
        colormap : str
            one of the colormaps in vispy
        """
        if color is not None and len(color) == 3:
            color = r_[array(color), 1.]  # make sure it's an array

        if values is not None:
            if limits_c is None:
                limits_c = nanmin(values), nanmax(values)

            norm_values = normalize(values, *limits_c)

            cm = get_colormap(colormap)
            vertex_colors = cm[norm_values].rgba

            hasnan = isnan(vertex_colors).all(axis=1)
            vertex_colors[hasnan, :] = color

        if vertex_colors is not None:
            color = None

        meshdata = MeshData(vertices=surf.vert, faces=surf.tri,
                            vertex_colors=vertex_colors)
        mesh = SimpleMesh(meshdata, color)

        self._plt.view.add(mesh)

        surf_center = mean(surf.vert, axis=0)
        if surf_center[0] < 0:
            azimuth = 270
        else:
            azimuth = 90

        self._plt.view.camera.center = surf_center
        self._plt.view.camera.scale_factor = SCALE_FACTOR
        self._plt.view.camera.elevation = ELEVATION
        self._plt.view.camera.azimuth = azimuth
        self._plt.view.border_color = 'w'  # border around surf

        self._surf.append(mesh)

    def add_chan(self, chan, color=CHAN_COLOR, chan_colors=None,
                 values=None, limits_c=None, colormap='coolwarm',
                 shift=(0, 0, 0)):
        """Add channels to visualization

        Parameters
        ----------
        chan : instance of Channels
            channels to plot
        color : tuple
            3-, 4-element tuple, representing RGB and alpha, between 0 and 1
        chan_colors : ndarray
            array with colors for each channel
        values : ndarray
            array with values for each channel
        limits_c : tuple of 2 floats, optional
            min and max values to normalize the color
        colormap : str
            one of the colormaps in vispy
        shift : tuple of 3 floats
            shift all electrodes by this amount (unit and order depend on
            xyz coordinates of the electrodes)
        """
        # larger if colors are meaningful
        if values is not None:
            meshdata = create_sphere(radius=3, method='ico')
        else:
            meshdata = create_sphere(radius=1.5, method='ico')

        chan_colors, limits = _prepare_chan_colors(color, chan_colors, values,
                                                   limits_c, colormap)

        # store values in case we update the values
        self._chan_colormap = colormap
        self._chan_limits_c = limits

        for i, one_chan in enumerate(chan.chan):
            if chan_colors.ndim == 2:
                chan_color = chan_colors[i, :]
            else:
                chan_color = chan_colors

            mesh = SimpleMesh(meshdata, chan_color)
            mesh.transform = STTransform(translate=one_chan.xyz + shift)
            self._plt.view.add(mesh)
            self._chan.append(mesh)

    def update_chan(self, color=CHAN_COLOR, chan_colors=None, values=None):
        """Update values or colors for channels. To change colormap or limits,
        you need to create a new figure.

        Parameters
        ----------
        color : tuple
            3-, 4-element tuple, representing RGB and alpha, between 0 and 1
        chan_colors : ndarray
            array with colors for each channel
        values : ndarray
            array with values for each channel
        """
        chan_colors, _ = _prepare_chan_colors(color, chan_colors, values,
                                              self._chan_limits_c,
                                              self._chan_colormap)

        for i, chan in enumerate(self._chan):
            if chan_colors.ndim == 2:
                chan_color = chan_colors[i, :]
            else:
                chan_color = chan_colors
            chan.update_color(chan_color)

        self._plt.update()


def _prepare_chan_colors(color, chan_colors, values, limits_c, colormap):
    """Return colors for all the channels based on various inputs.

    Parameters
    ----------
    color : tuple
        3-, 4-element tuple, representing RGB and alpha, between 0 and 1
    chan_colors : ndarray
        array with colors for each channel
    values : ndarray
        array with values for each channel
    limits_c : tuple of 2 floats, optional
        min and max values to normalize the color
    colormap : str
        one of the colormaps in vispy

    Returns
    -------
    1d / 2d array
        colors for all the channels or for each channel individually
    tuple of two float or None
        limits for the values
    """
    if values is not None:

        if limits_c is None:
            limits_c = min(values), max(values)

        norm_values = normalize(values, *limits_c)

        cm = get_colormap(colormap)
        colors = cm[norm_values].rgba

    else:
        colors = array(color)   # make sure it's an array
        if len(colors) == 3:
            colors = r_[colors, 1.]

    return colors, limits_c



"""Module to plot all the elements as flat images.
"""
from numpy import nanmax, nanmin

from .base import COLORMAP, Viz


class Viz2(Viz):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def add_data(self, data, trial=0, limits_c=None, colormap=COLORMAP,
                 **kwargs):
        """
        Parameters
        ----------
        data : any instance of DataType
            Duck-typing should help
        trial : int
            index of the trial to plot
        limits_c : tuple, optional
            limits on the z-axis (if unspecified, it's the max across subplots)
        """
        kwargs.update({'trial': trial})

        dat = data(**kwargs)

        if limits_c is None:
            max_c = nanmax(dat)
            min_c = nanmin(dat)
        else:
            min_c, max_c = limits_c

        print(min_c)
        print(max_c)

        plt = self._fig[0, 0]
        plt.image(dat.T, clim=(min_c, max_c), cmap=colormap)
        plt.view.camera.aspect = None
        plt.xaxis.axis.domain = (data.axis['time'][trial][0],
                                 data.axis['time'][trial][-1])
        self._plt = plt



"""Module with helper functions for plotting
"""
from vispy.app import Application
app = Application('pyqt5')

from numpy import array, clip, float32, r_
from vispy.scene.visuals import create_visual_node
from vispy.gloo import VertexBuffer
from vispy.io.image import _make_png
from vispy.plot import Fig
from vispy.visuals import Visual
from vispy.gloo.wrappers import read_pixels

DPI = 90
INCH_IN_MM = 25.4
COLORMAP = 'coolwarm'


vertex_shader = """
varying vec3 normal_vec;
varying vec4 color_vec;

void main() {
    normal_vec = $normal;
    color_vec = $color;
    gl_Position = $transform(vec4($position, 1));
}
"""

fragment_shader = """
varying vec3 normal_vec;
varying vec4 color_vec;

void main() {
    float p = dot(normal_vec, normalize(vec3($light_vec)));

    vec4 color = color_vec;
    color.rgb = color.rgb * (0.2 + 0.8 * abs(p));

    gl_FragColor = color;
}
"""


class Viz():

    def __init__(self, show=True, size_mm=None, dpi=None):

        kwargs = {}
        if dpi is None:
            kwargs['dpi'] = DPI
        else:
            kwargs['dpi'] = dpi
        if size_mm is not None:
            kwargs['size'] = (array(size_mm) / INCH_IN_MM * kwargs['dpi']).astype(int)

        self._fig = Fig(show=show, **kwargs)

    def _repr_png_(self):
        """This is used by ipython to plot inline.
        """
        try:
            region = (0, 0,  # TODO: how to get these values
                      int(self._plt.view.bounds(0)[1]),
                      int(self._plt.view.bounds(1)[1]))
        except AttributeError:
            region = None

        self._fig.show(True)
        app.process_events()
        img = read_pixels(viewport=region)  # self._fig.render(region=region)
        self._fig.show(False)
        return bytes(_make_png(img))

    def save(self, png_file):
        """Save png to disk.

        Parameters
        ----------
        png_file : path to file
            file to write to

        Notes
        -----
        It relies on _repr_png_, so fix issues there.
        """
        with open(png_file, 'wb') as f:
            f.write(self._repr_png_())


def normalize(x, min_value, max_value):
    """Normalize value between min and max values.
    It also clips the values, so that you cannot have values higher or lower
    than 0 - 1."""
    x = (x - min_value) / (max_value - min_value)
    return clip(x, 0, 1)


class SimpleMeshVisual(Visual):
    """Simple Visual for Mesh, with light coming from left and right.

    Parameters
    ----------
    meshdata : instance of vispy.geometry.MeshData
        the meshdata, with or without colors for each vertex
    color : 3- or 4-element tuple
        color for the whole mesh (color for each vertex should be in meshdata)
    light_vec : 3-element tuple
        light direction, only the axis, because light comes from both directions

    Notes
    -----
    I decided not to use the mesh visual that comes with vispy because I don't
    like the mirror reflection, which is really strong. Also, here I can
    control the parameters. The hard part is that it relies on glsl above.
    """
    def __init__(self, meshdata, color=None, light_vec=(1, 0, 0)):
        Visual.__init__(self, vertex_shader, fragment_shader)

        v = meshdata.get_vertices(indexed='faces').astype(float32)
        self._vertices = VertexBuffer(v)

        v_norm = meshdata.get_vertex_normals(indexed='faces').astype(float32)
        self._normals = VertexBuffer(v_norm)

        if color is not None:
            if len(color) == 3:
                color = r_[array(color), 1.]
            self._colors = color

        else:
            v_col = meshdata.get_vertex_colors(indexed='faces').astype(float32)
            self._colors = VertexBuffer(v_col)

        self._light_vec = light_vec

        self._draw_mode = 'triangles'
        self.set_gl_state('opaque', depth_test=True, cull_face=True)

    def update_color(self, color):
        self._colors = color

    def _prepare_transforms(self, view):
        view.view_program.vert['transform'] = view.get_transform()

    def _prepare_draw(self, view):
        self.shared_program.vert['position'] = self._vertices
        self.shared_program.vert['normal'] = self._normals
        self.shared_program.vert['color'] = self._colors
        self.shared_program.frag['light_vec'] = self._light_vec

SimpleMesh = create_visual_node(SimpleMeshVisual)



"""Module reads and writes header and data for EDF data.

Poor man's version of
https://github.com/breuderink/eegtools/blob/master/eegtools/io/edfplus.py

Values are slightly different from those computed by FieldTrip, however they
are identical to those computed by Biosig and EDFBrowser. The difference is due
to the calibration.

"""
from logging import getLogger
lg = getLogger(__name__)

from datetime import datetime, timedelta
from math import floor
from re import findall
from struct import pack

from numpy import empty, asarray, fromstring, iinfo, abs, max

EDF_FORMAT = 'int16'  # by definition
edf_iinfo = iinfo(EDF_FORMAT)
DIGITAL_MAX = edf_iinfo.max
DIGITAL_MIN = -1 * edf_iinfo.max  # so that digital 0 = physical 0


def _assert_all_the_same(items):
    """Check that all the items in a list are the same.

    """
    assert all(items[0] == x for x in items)
    
def _check_all_the_same(items):
    """Same as _assert_all_the_same, but instead returns bool."""
    return all(items[0] == x for x in items)


class Edf:
    """Provide class EDF, which can be used to read the header and the data.

    Parameters
    ----------
    edffile : str
        Full path for the EDF file

    Attributes
    ----------
    h : dict
        disorganized header information

    """

    def __init__(self, edffile):
        if isinstance(edffile, str):
            self.filename = edffile
            self._read_hdr()

    def _read_hdr(self):
        """Read header from EDF file.

        It only reads the header for internal purposes and adds a hdr.

        """

        with open(self.filename, 'rb') as f:

            hdr = {}
            assert f.tell() == 0
            assert f.read(8) == b'0       '

            # recording info
            hdr['subject_id'] = f.read(80).decode('utf-8').strip()
            hdr['recording_id'] = f.read(80).decode('utf-8').strip()

            # parse timestamp
            (day, month, year) = [int(x) for x in findall('(\d+)',
                                  f.read(8).decode('utf-8'))]
            (hour, minute, sec) = [int(x) for x in findall('(\d+)',
                                   f.read(8).decode('utf-8'))]
            hdr['start_time'] = datetime(year + 2000, month, day, hour, minute,
                                         sec)

            # misc
            hdr['header_n_bytes'] = int(f.read(8))
            f.seek(44, 1)  # reserved for EDF+
            hdr['n_records'] = int(f.read(8))
            hdr['record_length'] = float(f.read(8))  # in seconds
            nchannels = hdr['n_channels'] = int(f.read(4))

            # read channel info
            channels = range(hdr['n_channels'])
            hdr['label'] = [f.read(16).decode('utf-8').strip() for n in
                            channels]
            hdr['transducer'] = [f.read(80).decode('utf-8').strip()
                                 for n in channels]
            hdr['physical_dim'] = [f.read(8).decode('utf-8').strip() for n in
                                   channels]
            hdr['physical_min'] = asarray([float(f.read(8))
                                           for n in channels])
            hdr['physical_max'] = asarray([float(f.read(8))
                                           for n in channels])
            hdr['digital_min'] = asarray([float(f.read(8)) for n in channels])
            hdr['digital_max'] = asarray([float(f.read(8)) for n in channels])
            hdr['prefiltering'] = [f.read(80).decode('utf-8').strip()
                                   for n in channels]
            hdr['n_samples_per_record'] = [int(f.read(8)) for n in channels]
            f.seek(32 * nchannels, 1)  # reserved

            assert f.tell() == hdr['header_n_bytes']

            self.hdr = hdr

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            additional information taken directly from the header

        """

        subj_id = self.hdr['subject_id']
        start_time = self.hdr['start_time']
        if not _check_all_the_same(self.hdr['n_samples_per_record']):
            lg.warning('Not all channels contain the same number of samples. '
                       'Avoid loading these channels together to prevent '
                       'misalignment.')
        s_freq = (max(self.hdr['n_samples_per_record']) /
                  self.hdr['record_length'])
        chan_name = self.hdr['label']
        n_samples = (max(self.hdr['n_samples_per_record']) *
                     self.hdr['n_records'])

        return subj_id, start_time, s_freq, chan_name, n_samples, self.hdr

    def _read_dat(self, i_chan, begsam, endsam):
        """Read raw data from a single EDF channel.

        Reads only one channel at the time. Very initial implementation, very
        simple and probably not very fast

        Parameters
        ----------
        i_chan : int
            index of the channel to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A vector with the data as written on file, in 16-bit precision

        """

        assert begsam < endsam

        begsam = float(begsam)
        endsam = float(endsam)

        n_sam_rec = self.hdr['n_samples_per_record']

        begrec = int(floor(begsam / n_sam_rec[i_chan]))
        begsam_rec = int(begsam % n_sam_rec[i_chan])

        endrec = int(floor(endsam / n_sam_rec[i_chan]))
        endsam_rec = int(endsam % n_sam_rec[i_chan])

        dat = empty(shape=(int(endsam) - int(begsam)), dtype='int16')
        i_dat = 0

        with open(self.filename, 'rb') as f:
            for rec in range(begrec, endrec + 1):
                if rec == begrec:
                    begpos_rec = begsam_rec
                else:
                    begpos_rec = 0

                if rec == endrec:
                    endpos_rec = endsam_rec
                else:
                    endpos_rec = n_sam_rec[i_chan]

                begpos = begpos_rec + sum(n_sam_rec) * rec + sum(
                    n_sam_rec[:i_chan])
                endpos = endpos_rec + sum(n_sam_rec) * rec + sum(
                    n_sam_rec[:i_chan])

                f.seek(begpos * 2 + self.hdr['header_n_bytes'])
                samples = f.read(2 * (endpos - begpos))

                i_dat_end = i_dat + endpos - begpos
                dat[i_dat:i_dat_end] = fromstring(samples, dtype='<i2')
                i_dat = i_dat_end

        return dat

    def return_dat(self, chan, begsam, endsam):
        """Read data from an EDF file.

        Reads channel by channel, and adjusts the values by calibration.

        Parameters
        ----------
        chan : list of str
            index (indices) of the channels to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A 2d matrix, where the first dimension is the channels and the
            second dimension are the samples.

        """
        hdr = self.hdr
        dig_min = hdr['digital_min']
        phys_min = hdr['physical_min']
        phys_range = hdr['physical_max'] - hdr['physical_min']
        dig_range = hdr['digital_max'] - hdr['digital_min']
        assert all(phys_range > 0)
        assert all(dig_range > 0)
        gain = phys_range / dig_range

        dat = empty(shape=(len(chan), endsam - begsam), dtype='float64')

        for i, i_chan in enumerate(chan):
            d = self._read_dat(i_chan, begsam, endsam).astype('float64')
            dat[i, :] = (d - dig_min[i_chan]) * gain[i_chan] + phys_min[i_chan]

        return dat

    def return_markers(self):
        """"""
        return []


def write_edf(data, filename, physical_max=1000):
    """Export data to FieldTrip.

    Parameters
    ----------
    data : instance of ChanTime
        data with only one trial
    filename : path to file
        file to export to (include '.mat')
    physical_max : int
        values above this parameter will be considered saturated (and also
        those that are too negative). This parameter defines the precision.

    Notes
    -----
    Data is always recorded as 2 Byte int (which is 'int16'), so precision is
    limited. You can control the precision with physical_max. To get the
    precision:

    >>> precision = physical_max / DIGITAL_MAX

    where DIGITAL_MAX is 32767.
    """
    if data.start_time is None:
        raise ValueError('Data should contain a valid start_time (as datetime)')
    start_time = data.start_time + timedelta(seconds=data.axis['time'][0][0])

    if physical_max is None:
        physical_max = max(abs(data.data[0]))

    precision = physical_max / DIGITAL_MAX
    lg.info('Data exported to EDF will have precision ' + str(precision))

    physical_min = -1 * physical_max
    dat = data.data[0] / physical_max * DIGITAL_MAX
    dat = dat.astype(EDF_FORMAT)
    dat[dat > DIGITAL_MAX] = DIGITAL_MAX
    dat[dat < DIGITAL_MIN] = DIGITAL_MIN

    with open(filename, 'wb') as f:
        f.write('{:<8}'.format(0).encode('ascii'))
        f.write('{:<80}'.format('X X X X').encode('ascii'))  # subject_id
        f.write('{:<80}'.format('Startdate X X X X').encode('ascii'))
        f.write(start_time.strftime('%d.%m.%y').encode('ascii'))
        f.write(start_time.strftime('%H.%M.%S').encode('ascii'))

        n_smp = data.data[0].shape[1]
        s_freq = int(data.s_freq)
        n_records = n_smp // s_freq  # floor
        record_length = 1
        n_channels = data.number_of('chan')[0]

        header_n_bytes = 256 + 256 * n_channels
        f.write('{:<8d}'.format(header_n_bytes).encode('ascii'))
        f.write((' ' * 44).encode('ascii'))  # reserved for EDF+

        f.write('{:<8}'.format(n_records).encode('ascii'))
        f.write('{:<8d}'.format(record_length).encode('ascii'))
        f.write('{:<4}'.format(n_channels).encode('ascii'))

        for chan in data.axis['chan'][0]:
            f.write('{:<16}'.format(chan).encode('ascii'))  # label
        for _ in range(n_channels):
            f.write(('{:<80}').format('').encode('ascii'))  # tranducer
        for _ in range(n_channels):
            f.write('{:<8}'.format('uV').encode('ascii'))  # physical_dim
        for _ in range(n_channels):
            f.write('{:<8}'.format(physical_min).encode('ascii'))
        for _ in range(n_channels):
            f.write('{:<8}'.format(physical_max).encode('ascii'))
        for _ in range(n_channels):
            f.write('{:<8}'.format(DIGITAL_MIN).encode('ascii'))
        for _ in range(n_channels):
            f.write('{:<8}'.format(DIGITAL_MAX).encode('ascii'))
        for _ in range(n_channels):
            f.write('{:<80}'.format('').encode('ascii'))  # prefiltering
        for _ in range(n_channels):
            f.write('{:<8d}'.format(s_freq).encode('ascii'))  # n_smp in record
        for _ in range(n_channels):
            f.write((' ' * 32).encode('ascii'))

        l = s_freq * n_channels  # length of one record
        for i in range(n_records):
            i0 = i * s_freq
            i1 = i0 + s_freq
            x = dat[:, i0:i1].flatten(order='C')  # assumes it's ChanTimeData
            f.write(pack('<' + 'h' * l, *x))



"""Module to detect spindles.

"""
from logging import getLogger
from numpy import (absolute, arange, argmax, asarray, concatenate, cos, diff, 
                   exp, empty, floor, hstack, insert, invert, linspace, 
                   logical_or, mean, median, nan, ones, pi, ptp, sqrt, square, 
                   std, vstack, where, zeros)
from scipy.ndimage.filters import gaussian_filter
from scipy.signal import (argrelmax, butter, cheby2, filtfilt, fftconvolve,
                          hilbert, periodogram)

from ..graphoelement import Spindles

lg = getLogger(__name__)
MAX_FREQUENCY_OF_INTEREST = 50
MAX_DURATION = 5


class DetectSpindle:
    """Design spindle detection on a single channel.

    Parameters
    ----------
    method : str
        one of the predefined methods (see below 'housestyle')
    frequency : tuple of float
        low and high frequency of spindle band
    duration : tuple of float
        min and max duration of spindles

    """
    def __init__(self, method='UCSD', frequency=None, duration=None, 
                 merge=True):

        if frequency is None:
            frequency = (11, 18)
        if duration is None:
            duration = (0.5, 2)

        self.method = method
        self.frequency = frequency
        self.duration = duration
        self.merge = merge
        self.min_interval = 0

        self.power_peaks = 'interval'

        if method == 'Ferrarelli2007':
            self.det_butter = {'order': 4,
                               'freq': self.frequency,
                               }
            self.det_wavelet = {'sd': None}
            self.det_thresh = 8
            self.sel_thresh = 2
            self.moving_rms = {'dur': None}
            self.smooth = {'dur': None}

        elif method == 'Nir2011':
            self.det_butter = {'order': 4,
                               'freq': self.frequency,
                               }
            self.det_wavelet = {'sd': None}
            self.det_thresh = 3
            self.sel_thresh = 1  
            self.min_interval = 1
            self.moving_rms = {'dur': None}
            self.smooth = {'dur': .4} # is in fact sigma

        elif method == 'Wamsley2012':
            self.det_wavelet = {'f0': mean(self.frequency),
                                'sd': .5,
                                'dur': 2,
                                }
            self.det_thresh = 4.5
            self.sel_thresh = None  # necessary for gui/detect
            self.moving_rms = {'dur': None}
            self.smooth = {'dur': .1}

        elif method == 'UCSD':
            self.det_wavelet = {'freqs': arange(frequency[0],
                                                frequency[1] + .5, .5),
                                'dur': 1,
                                'width': .5,
                                'win': .5,
                                'sd': None
                                }
            self.det_thresh = 2  # wavelet_peak_thresh
            self.sel_wavelet = {'freqs': arange(frequency[0],
                                                frequency[1] + .5, .5),
                                'dur': 1,
                                'width': .2,
                                'win': .2,
                                }
            self.sel_thresh = 1
            self.ratio_thresh = .5
            self.moving_rms = {'dur': None}
            self.smooth = {'dur': None}
            
        elif method == 'Moelle2011':
            self.det_butter = {'order': 4,
                               'freq': self.frequency,
                               }
            self.det_wavelet = {'sd': None}
            self.det_thresh = 1.5
            self.sel_thresh = None
            self.moving_rms = {'dur': .2}
            self.smooth = {'dur': .2}
            
        else:
            raise ValueError('Unknown method')

    def __repr__(self):
        return ('detsp_{0}_{1:02}-{2:02}Hz_{3:04.1f}-{4:04.1f}s'
                ''.format(self.method, self.frequency[0], self.frequency[1],
                          self.duration[0], self.duration[1]))

    def __call__(self, data, make_plots=False):
        """Detect spindles on the data.

        Parameters
        ----------
        data : instance of Data
            data used for detection

        Returns
        -------
        instance of graphoelement.Spindles
            description of the detected spindles

        """
        spindle = Spindles()
        spindle.chan_name = data.axis['chan'][0]
        spindle.det_value = zeros(data.number_of('chan')[0])
        spindle.sel_value = zeros(data.number_of('chan')[0])
        spindle.density = zeros(data.number_of('chan')[0])

        all_spindles = []
        for i, chan in enumerate(data.axis['chan'][0]):
            lg.info('Detecting spindles on chan %s', chan)
            time = hstack(data.axis['time'])
            dat_orig = hstack(data(chan=chan))

            if self.method == 'Ferrarelli2007':
                sp_in_chan, values, density = detect_Ferrarelli2007(dat_orig,
                                                           data.s_freq, time,
                                                           self)

            elif self.method == 'Nir2011':
                sp_in_chan, values, density = detect_Nir2011(dat_orig, 
                                                             data.s_freq,
                                                             time, self)

            elif self.method == 'Wamsley2012':
                sp_in_chan, values, density = detect_Wamsley2012(dat_orig, 
                                                                 data.s_freq,
                                                                 time, self)

            elif self.method == 'UCSD':
                sp_in_chan, values, density = detect_UCSD(dat_orig, 
                                                          data.s_freq, time,
                                                          self)
            elif self.method == 'Moelle2011':
                sp_in_chan, values, density = detect_Moelle2011(dat_orig, 
                                                                data.s_freq, 
                                                                time, self)
            else:
                raise ValueError('Unknown method')

            spindle.det_value[i] = values['det_value']
            spindle.sel_value[i] = values['sel_value']
            spindle.density[i] = density
                        
            for sp in sp_in_chan:
                sp.update({'chan': chan})
            
            all_spindles.extend(sp_in_chan)
            # end of loop over chan

        spindle.events = sorted(all_spindles, key=lambda x: x['start'])
        
        if self.merge and len(data.axis['chan'][0]) > 1:
            spindle.events = merge_close(spindle.events, self.min_interval)

        return spindle


def detect_Ferrarelli2007(dat_orig, s_freq, time, opts):
    """Spindle detection based on Ferrarelli et al. 2007.

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        'det_cheby2' : dict
            parameters for 'cheby2',
        'det_thresh' : float
            detection threshold
        'sel_thresh' : float
            selection threshold
        'duration' : tuple of float
            min and max duration of spindles

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value' with detection value, 'sel_value' with selection value
    float
        spindle density, per 30-s epoch

    Notes
    -----
    The original article does not specify a filter, but cheby2 seems the best
    to me.

    References
    ----------
    Ferrarelli, F. et al. Am. J. Psychiatry 164, 483-92 (2007).

    """
    dat_det = transform_signal(dat_orig, s_freq, 'butter', opts.det_butter)
    dat_det = transform_signal(dat_det, s_freq, 'hilbert')
    dat_det = transform_signal(dat_det, s_freq, 'abs')
    
    det_value = define_threshold(dat_det, s_freq, 'mean', opts.det_thresh)
    sel_value = define_threshold(dat_det, s_freq, 'mean', opts.sel_thresh)

    events = detect_events(dat_det, 'above_thresh', det_value)

    if events is not None:
        events = select_events(dat_det, events, 'above_thresh', sel_value)

        events = _merge_close(dat_det, events, time, opts.min_interval)

        events = within_duration(events, time, opts.duration)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        power_avgs = avg_power(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, power_avgs, dat_det, 
                                   dat_orig, time, s_freq)

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'det_value': det_value, 'sel_value': sel_value}
    lg.info(str(values))

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_Moelle2011(dat_orig, s_freq, time, opts):
    """Spindle detection based on Moelle et al. 2011
    
    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    opts : instance of 'DetectSpindle'
        'det_butter' : dict
            parameters for 'butter',
        'moving_rms' : dict
            parameters for 'moving_rms'
        'smooth' : dict
            parameters for 'moving_avg'
        'det_thresh' : float
            detection threshold
        'sel_thresh' : nan
            not used, but keep it for consistency with the other methods
        'duration' : tuple of float
            min and max duration of spindles
            
    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value' with detection value, 'sel_value' with nan
    float
        spindle density, per 30-s epoch
        
    Notes
    -----
    The original article does not specify a filter, but butter seems the best
    to me.
        
    References
    ----------
    Moelle, M. et al. Sleep 34, 1411-21 (2011).
    
    """
    dat_det = transform_signal(dat_orig, s_freq, 'butter', opts.det_butter)
    dat_det = transform_signal(dat_det, s_freq, 'moving_rms', opts.moving_rms)
    dat_det = transform_signal(dat_det, s_freq, 'moving_avg', opts.smooth)
    
    det_value = define_threshold(dat_det, s_freq, 'mean+std', opts.det_thresh)
    
    events = detect_events(dat_det, 'above_thresh', det_value)
    
    if events is not None:
        events = _merge_close(dat_det, events, time, opts.min_interval)
        
        events = within_duration(events, time, opts.duration)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        power_avgs = avg_power(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, power_avgs, dat_det, 
                                   dat_orig, time, s_freq)

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'det_value': det_value, 'sel_value': nan}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_Nir2011(dat_orig, s_freq, time, opts):
    """Spindle detection based on Nir et al. 2011

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        'det_butter' : dict
            parameters for 'butter',
        'smooth' : dict
            parameters for 'gaussian'
        'det_thresh' : float
            detection threshold
        'sel_thresh' : float
            selection threshold
        'min_interval' : float
            minimum interval between consecutive events
        'duration' : tuple of float
            min and max duration of spindles

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value' with detection value, 'sel_value' with selection value
    float
        spindle density, per 30-s epoch

    Notes
    -----
    This paper also selects channels carefully:
    'First, the channels with spindle activity in NREM sleep were
    chosen for further analysis.'

    'Third, those channels, in which an increase in spectral power
    within the detected events was restricted to the spindle-frequency
    range (10-16 Hz) rather than broadband.'

    References
    ----------
    Nir, Y. et al. Neuron 70, 153-69 (2011).

    """
    dat_det = transform_signal(dat_orig, s_freq, 'butter', opts.det_butter)
    dat_det = transform_signal(dat_det, s_freq, 'hilbert')
    dat_det = transform_signal(dat_det, s_freq, 'abs')
    dat_det = transform_signal(dat_det, s_freq, 'gaussian', opts.smooth)
    
    det_value = define_threshold(dat_det, s_freq, 'mean+std', opts.det_thresh)
    sel_value = define_threshold(dat_det, s_freq, 'mean+std', opts.sel_thresh)

    events = detect_events(dat_det, 'above_thresh', det_value)

    if events is not None:
        events = select_events(dat_det, events, 'above_thresh', sel_value)
        
        events = _merge_close(dat_det, events, time, opts.min_interval)

        events = within_duration(events, time, opts.duration)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        power_avgs = avg_power(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, power_avgs, dat_det, 
                                   dat_orig, time, s_freq)

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'det_value': det_value, 'sel_value': sel_value}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_Wamsley2012(dat_orig, s_freq, time, opts):
    """Spindle detection based on Wamsley et al. 2012

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        'det_wavelet' : dict
            parameters for 'morlet',
        'smooth' : dict
            parameters for 'moving_avg'
        'det_thresh' : float
            detection threshold
        'sel_thresh' : nan
            not used, but keep it for consistency with the other methods
        'duration' : tuple of float
            min and max duration of spindles

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value' with detection value, 'sel_value' is nan (for consistency
        with other methods)
    float
        spindle density, per 30-s epoch

    References
    ----------
    Wamsley, E. J. et al. Biol. Psychiatry 71, 154-61 (2012).

    """
    dat_det = transform_signal(dat_orig, s_freq, 'morlet', opts.det_wavelet)
    dat_det = transform_signal(dat_det, s_freq, 'abs')
    dat_det = transform_signal(dat_det, s_freq, 'moving_avg', opts.smooth)
    
    det_value = define_threshold(dat_det, s_freq, 'mean', opts.det_thresh)

    events = detect_events(dat_det, 'above_thresh', det_value)

    if events is not None:
        events = _merge_close(dat_det, events, time, opts.min_interval)
        
        events = within_duration(events, time, opts.duration)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        power_avgs = avg_power(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, power_avgs, dat_det, 
                                   dat_orig, time, s_freq)

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'det_value': det_value, 'sel_value': nan}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_UCSD(dat_orig, s_freq, time, opts):
    """Spindle detection based on the UCSD method

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        det_wavelet : dict
            parameters for 'wavelet_real',
        det_thres' : float
            detection threshold
        sel_thresh : float
            selection threshold
        duration : tuple of float
            min and max duration of spindles
        frequency : tuple of float
            low and high frequency of spindle band (for power ratio)
        ratio_thresh : float
            ratio between power inside and outside spindle band to accept them

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value' with detection value, 'sel_value' is selection value
    float
        spindle density, per 30-s epoch

    """
    dat_det = transform_signal(dat_orig, s_freq, 'wavelet_real',
                               opts.det_wavelet)

    det_value = define_threshold(dat_det, s_freq, 'median+std',
                                 opts.det_thresh)

    events = detect_events(dat_det, 'maxima', det_value)

    dat_sel = transform_signal(dat_orig, s_freq, 'wavelet_real',
                               opts.sel_wavelet)
    sel_value = define_threshold(dat_sel, s_freq, 'median+std',
                                 opts.sel_thresh)
    events = select_events(dat_sel, events, 'above_thresh', sel_value)

    events = _merge_close(dat_det, events, time, opts.min_interval)

    events = within_duration(events, time, opts.duration)

    events = power_ratio(events, dat_orig, s_freq, opts.frequency,
                         opts.ratio_thresh)

    power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
    power_avgs = avg_power(events, dat_orig, s_freq, opts.frequency)
    sp_in_chan = make_spindles(events, power_peaks, power_avgs, dat_det, 
                               dat_orig, time, s_freq)

    values = {'det_value': det_value, 'sel_value': sel_value}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def transform_signal(dat, s_freq, method, method_opt=None):
    """Transform the data using different methods.

    Parameters
    ----------
    dat : ndarray (dtype='float')
        vector with all the data for one channel
    s_freq : float
        sampling frequency
    method : str
        one of 'cheby2', 'butter', 'morlet', 'morlet_real', 'hilbert', 'abs',
        'moving_avg', 'gaussian'
    method_opt : dict
        depends on methods

    Returns
    -------
    ndarray (dtype='float')
        vector with all the data for one channel

    Notes
    -----
    Wavelets pass only absolute values already, it does not make sense to store
    the complex values.

    Methods
    -------
    cheby2 has parameters:
        freq : tuple of float
            high and low values for bandpass
        order : int
            filter order

    butter has parameters:
        freq : tuple of float
            high and low values for bandpass
        order : int
            filter order

    morlet has parameters:
        f0 : float
            center frequency in Hz
        sd : float
            standard deviation of frequency
        dur : float
            window length in number of standard deviations

    morlet_real has parameters:
        freqs : ndarray
            vector of wavelet frequencies for spindle detection
        dur : float
            duration of the wavelet (sec)
        width : float
            wavelet width
        win : float
            moving average window length (sec) of wavelet convolution

    moving_avg has parameters:
        dur : float
            duration of the window (sec)
            
    moving_rms has parameters:
        dur : float
            duration of the window (sec)
            
    gaussian has parameters:
        dur : float
            standard deviation of the Gaussian kernel, aka sigma (sec)

    """
    if 'cheby2' == method:
        freq = method_opt['freq']
        N = method_opt['order']

        Rs = 80
        nyquist = s_freq / 2
        Wn = asarray(freq) / nyquist
        b, a = cheby2(N, Rs, Wn, btype='bandpass')
        dat = filtfilt(b, a, dat)

    if 'butter' == method:
        freq = method_opt['freq']
        N = method_opt['order']

        nyquist = s_freq / 2
        Wn = asarray(freq) / nyquist
        b, a = butter(N, Wn, btype='bandpass')
        dat = filtfilt(b, a, dat)

    if 'morlet' == method:
        f0 = method_opt['f0']
        sd = method_opt['sd']
        dur = method_opt['dur']

        wm = _wmorlet(f0, sd, s_freq, dur)
        dat = absolute(fftconvolve(dat, wm, mode='same'))

    if 'wavelet_real' == method:
        freqs = method_opt['freqs']
        dur = method_opt['dur']
        width = method_opt['width']
        win = method_opt['win'] * s_freq

        wm = _realwavelets(s_freq, freqs, dur, width)
        tfr = empty((dat.shape[0], wm.shape[0]))
        for i, one_wm in enumerate(wm):
            x = abs(fftconvolve(dat, one_wm, mode='same'))
            tfr[:, i] = fftconvolve(x, tukeywin(win), mode='same')
        dat = mean(tfr, axis=1)

    if 'hilbert' == method:
        dat = hilbert(dat)

    if 'abs' == method:
        dat = absolute(dat)

    if 'moving_avg' == method:
        dur = method_opt['dur']

        flat = ones(int(dur * s_freq))
        dat = fftconvolve(dat, flat / sum(flat), mode='same')
        
    if 'moving_rms' == method:
        dur = method_opt['dur']
        halfdur = int(floor(s_freq * dur / 2))
        ldat = len(dat)
        rms = zeros((ldat))
        
        for i in range(ldat):
            rms[i] = sqrt(mean(square(
                    dat[max(0, i - halfdur):min(ldat, i + halfdur)]
                    )))
        dat = rms
        
    if 'gaussian' == method:
        sigma = method_opt['dur']
        
        dat = gaussian_filter(dat, sigma)

    return dat


def define_threshold(dat, s_freq, method, value):
    """Return the value of the threshold based on relative values.

    Parameters
    ----------
    dat : ndarray (dtype='float')
        vector with the data after selection-transformation
    s_freq : float
        sampling frequency
    method : str
        one of 'mean', 'median', 'std', 'mean+std', 'median+std'
    value : float
        value to multiply the values for

    Returns
    -------
    float
        threshold in useful units.

    """
    if method == 'mean':
        value = value * mean(dat)
    elif method == 'median':
        value = value * median(dat)
    elif method == 'std':
        value = value * std(dat)
    elif method == 'mean+std':
        value = mean(dat) + value * std(dat)
    elif method == 'median+std':
        value = median(dat) + value * std(dat)

    return value


def detect_events(dat, method, value=None):
    """Detect events using 'above_thresh', 'below_thresh' or 
    'maxima' method.

    Parameters
    ----------
    dat : ndarray (dtype='float')
        vector with the data after transformation
    method : str
        'above_thresh', 'below_thresh' or 'maxima'
    value : float
        for 'threshold's, it's the value of threshold for the event detection
        for 'maxima', it's the distance in s from the peak to find a minimum

    Returns
    -------
    ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples

    """
    if method == 'above_thresh':
        above_det = dat >= value
        detected = _detect_start_end(above_det)
        
    if method == 'below_thresh':
        below_det = dat < value
        lg.info('values below zero is ' + str(below_det.any()))
        detected = _detect_start_end(below_det)
    
    if method in ['above_thresh', 'below_thresh']:
        if detected is None:
            return None

        # add the location of the peak in the middle
        detected = insert(detected, 1, 0, axis=1)
        for i in detected:
            i[1] = i[0] + argmax(dat[i[0]:i[2]])

    if method == 'maxima':
        peaks = argrelmax(dat)[0]
        detected = vstack((peaks, peaks, peaks)).T

        if value is not None:
            detected = detected[dat[peaks] > value, :]

    return detected


def select_events(dat, detected, method, value):
    """Select start sample and end sample of the events.

    Parameters
    ----------
    dat : ndarray (dtype='float')
        vector with the data after selection-transformation
    detected : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    method : str
        'above_thresh', 'below_thresh'
    value : float
        for 'threshold', it's the value of threshold for the spindle selection.

    Returns
    -------
    ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples

    """
    if method == 'above_thresh':
        above_sel = dat >= value
        detected = _select_period(detected, above_sel)
    elif method == 'below_thresh':
        below_sel = dat <= value
        detected = _select_period(detected, below_sel)

    return detected


def merge_close(events, min_interval):
    """Merge events that are separated by a less than a minimum interval.
    When merging, longer event is extended and shorter one is discarded.
    
    Parameters
    ----------
    events : list of dict
        events with 'start' and 'end' times, from one or several channels.
        **Events must be sorted by their start time.**
    min_interval : float
        minimum delay between consecutive events, in seconds
        
    Returns
    -------
    list of dict
        original events list with close events merged. Note that in a merger,
        info (chan, peak, etc.) from the longer of the 2 events is kept.
    """
    half_iv = min_interval / 2    
    merged = []
    
    for higher in events:

        if not merged:
            merged.append(higher)
            
        else:
            lower = merged[-1]

            if higher['start'] - half_iv <= lower['end'] + half_iv:
                
                if higher['end'] - higher['start'] > \
                lower['end'] - lower['start']:
                    start = min(lower['start'], higher['start'])
                    higher.update({'start': start})
                    merged[-1] = higher

                else:
                    end = max(lower['end'], higher['end'])
                    merged[-1].update({'end': end})
                             
            else:
                merged.append(higher)
                
    return merged


def within_duration(events, time, limits):
    """Check whether event is within time limits.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    time : ndarray (dtype='float')
        vector with time points
    limits : tuple of float
        low and high limit for spindle duration

    Returns
    -------
    ndarray (dtype='int')
        N x 3 matrix with start , peak, end samples

    """
    min_dur = time[events[:, 2] - 1] - time[events[:, 0]] >= limits[0]
    max_dur = time[events[:, 2] - 1] - time[events[:, 0]] <= limits[1]

    return events[min_dur & max_dur, :]


def power_ratio(events, dat, s_freq, limits, ratio_thresh):
    """Estimate the ratio in power between spindle band and lower frequencies.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    dat : ndarray (dtype='float')
        vector with the original data
    s_freq : float
        sampling frequency
    limits : tuple of float
        high and low frequencies for spindle band
    ratio_thresh : float
        ratio between spindle vs non-spindle amplitude

    Returns
    -------
    ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples

    Notes
    -----
    In the original matlab script, it uses amplitude, not power.

    """
    ratio = empty(events.shape[0])
    for i, one_event in enumerate(events):

        x0 = one_event[0]
        x1 = one_event[2]

        if x0 < 0 or x1 >= len(dat):
            ratio[i] = 0

        else:
            f, Pxx = periodogram(dat[x0:x1], s_freq, scaling='spectrum')
            Pxx = sqrt(Pxx)  # use amplitude

            freq_sp = (f >= limits[0]) & (f <= limits[1])
            freq_nonsp = (f <= limits[1])

            ratio[i] = mean(Pxx[freq_sp]) / mean(Pxx[freq_nonsp])

    events = events[ratio > ratio_thresh, :]

    return events


def peak_in_power(events, dat, s_freq, method, value=None):
    """Define peak in power of the signal.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    dat : ndarray (dtype='float')
        vector with the original data
    s_freq : float
        sampling frequency
    method : str or None
        'peak' or 'interval'. If None, values will be all NaN
    value : float
        size of the window around peak, or nothing (for 'interval')

    Returns
    -------
    ndarray (dtype='float')
        vector with peak frequency

    """
    dat = diff(dat)  # remove 1/f

    peak = empty(events.shape[0])
    peak.fill(nan)

    if method is not None:
        for i, one_event in enumerate(events):

            if method == 'peak':
                x0 = one_event[1] - value / 2 * s_freq
                x1 = one_event[1] + value / 2 * s_freq

            elif method == 'interval':
                x0 = one_event[0]
                x1 = one_event[2]

            if x0 < 0 or x1 >= len(dat):
                peak[i] = nan
            else:
                f, Pxx = periodogram(dat[x0:x1], s_freq)
                idx_peak = Pxx[f < MAX_FREQUENCY_OF_INTEREST].argmax()
                peak[i] = f[idx_peak]

    return peak


def avg_power(events, dat, s_freq, frequency):
    """Define average power of the signal within frequency band.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    dat : ndarray (dtype='float')
        vector with the original data
    s_freq : float
        sampling frequency
    frequency : tuple of float
        low and high frequency of spindle band, for window

    Returns
    -------
    ndarray (dtype='float')
        vector with avg power

    """
    dat = diff(dat)  # remove 1/f

    avg = empty(events.shape[0])
    avg.fill(nan)

    for i, one_event in enumerate(events):

        x0 = one_event[0]
        x1 = one_event[2]

        if x0 < 0 or x1 >= len(dat):
            avg[i] = nan
        else:
            sf, Pxx = periodogram(dat[x0:x1], s_freq)
            # find nearest frequencies in sf
            b0 = asarray([abs(x - frequency[0]) for x in sf]).argmin()
            b1 = asarray([abs(x - frequency[1]) for x in sf]).argmin()
            avg[i] = mean(Pxx[b0:b1])

    return avg


def make_spindles(events, power_peaks, power_avgs, dat_det, dat_orig, time, 
                  s_freq):
    """Create dict for each spindle, based on events of time points.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples, and peak frequency
    power_peaks : ndarray (dtype='float')
        peak in power spectrum for each event
    power_avgs : ndarray (dtype='float')
        average power in power spectrum for each event
    dat_det : ndarray (dtype='float')
        vector with the data after detection-transformation (to compute peak)
    dat_orig : ndarray (dtype='float')
        vector with the raw data on which detection was performed
    time : ndarray (dtype='float')
        vector with time points
    s_freq : float
        sampling frequency

    Returns
    -------
    list of dict
        list of all the spindles, with information about start_time, peak_time,
        end_time (s), peak_val (signal units), area_under_curve
        (signal units * s), peak_freq (Hz)

    """
    i, events = _remove_duplicate(events, dat_det)
    power_peaks = power_peaks[i]

    spindles = []
    for i, one_peak, one_pwr in zip(events, power_peaks, power_avgs):
        one_spindle = {'start': time[i[0]],
                       'end': time[i[2] - 1],
                       'peak_time': time[i[1]],
                       'peak_val': dat_det[i[1]],
                       'peak_val_orig': dat_orig[i[1]],
                       'dur': (i[2] - i[0]) / s_freq,
                       'area_under_curve': sum(dat_det[i[0]:i[2]]) / s_freq,
                       'rms': sqrt(mean(square(dat_orig[i[0]:i[2]]))),
                       'power': one_pwr,
                       'peak_freq': one_peak,
                       'ptp': ptp(dat_orig[i[0]:i[2]])
                       }
        spindles.append(one_spindle)

    return spindles


def _remove_duplicate(old_events, dat):
    """Remove duplicates from the events.

    Parameters
    ----------
    old_events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    dat : ndarray (dtype='float')
        vector with the data after detection-transformation (to compute peak)

    Returns
    -------
    ndarray (dtype='int')
        vector of indices of the events to keep
    ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples

    Notes
    -----
    old_events is assumed to be sorted. It only checks for the start time and
    end time. When two (or more) events have the same start time and the same
    end time, then it takes the largest peak.

    There is no tolerance, indices need to be identical.

    """
    diff_events = diff(old_events, axis=0)
    dupl = where((diff_events[:, 0] == 0) & (diff_events[:, 2] == 0))[0]
    dupl += 1  # more convenient, it copies old_event first and then compares

    n_nondupl_events = old_events.shape[0] - len(dupl)
    new_events = zeros((n_nondupl_events, old_events.shape[1]), dtype='int')
    if len(dupl):
        lg.debug('Removing ' + str(len(dupl)) + ' duplicate events')

    i = 0
    indices = []
    for i_old, one_old_event in enumerate(old_events):
        if i_old not in dupl:
            new_events[i, :] = one_old_event
            i += 1
            indices.append(i_old)
        else:
            peak_0 = new_events[i - 1, 1]
            peak_1 = one_old_event[1]
            if dat[peak_0] >= dat[peak_1]:
                new_events[i - 1, 1] = peak_0
            else:
                new_events[i - 1, 1] = peak_1

    return indices, new_events


def _detect_start_end(true_values):
    """From ndarray of bool values, return intervals of True values.

    Parameters
    ----------
    true_values : ndarray (dtype='bool')
        array with bool values

    Returns
    -------
    ndarray (dtype='int')
        N x 2 matrix with starting and ending times.

    """
    neg = zeros((1), dtype='bool')
    int_values = asarray(concatenate((neg, true_values, neg)), dtype='int')
    cross_threshold = diff(int_values)

    event_starts = where(cross_threshold == 1)[0]
    event_ends = where(cross_threshold == -1)[0]

    if len(event_starts):
        events = vstack((event_starts, event_ends)).T

    else:
        events = None

    return events


def _select_period(detected, true_values):
    """For the detected values, we check when it goes above/below the 
    selection.

    Parameters
    ----------
    detected : ndarray (dtype='int')
        N x 3 matrix with starting and ending times.
    true_values : ndarray (dtype='bool')
        array with bool values

    Returns
    -------
    ndarray (dtype='int')
        N x 2 matrix with starting and ending times, but these periods are
        usually larger than those of the input, because the selection window is
        usually more lenient (lower threshold) than the detection window.

    Notes
    -----
    Both start and end time points are inclusive (not python convention, but
    matlab convention) because these values are converted to time points later.

    """
    true_values = invert(true_values)

    for one_spindle in detected:
        # get the first time point when it goes above/below selection thres
        start_sel = where(true_values[:one_spindle[0]])[0]
        if start_sel.any():
            one_spindle[0] = start_sel[-1]

        # get the last time point when it stays above/below selection thres
        end_sel = where(true_values[one_spindle[2]:])[0] - 1
        if end_sel.any():
            one_spindle[2] += end_sel[0]

    return detected


def _merge_close(dat, events, time, min_interval):
    """Merge together events separated by less than a minimum interval.
    
    Parameters
    ----------
    dat : ndarray (dtype='float')
        vector with the data after selection-transformation
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    time : ndarray (dtype='float')
        vector with time points
    min_interval : float
        minimum delay between consecutive events, in seconds
        
    Returns
    -------
    ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    """    
    if min_interval == 0:
        return events
    
    no_merge = time[events[1:, 0] - 1] - time[events[:-1, 2]] >= min_interval
    
    if no_merge.any():
        begs = concatenate([[events[0, 0]], events[1:, 0][no_merge]])
        ends = concatenate([events[:-1, 2][no_merge], [events[-1, 2]]])  
        
        new_events = vstack((begs, ends)).T
    else:
        new_events = asarray[[events[0, 0], events[-1, 2]]]
    
    # add the location of the peak in the middle
    new_events = insert(new_events, 1, 0, axis=1)
    for i in new_events:
        i[1] = i[0] + argmax(dat[i[0]:i[2]])
    
    return new_events


def _wmorlet(f0, sd, sampling_rate, ns=5):
    """
    adapted from nitime

    returns a complex morlet wavelet in the time domain

    Parameters
    ----------
        f0 : center frequency
        sd : standard deviation of frequency
        sampling_rate : samplingrate
        ns : window length in number of standard deviations

    """
    st = 1. / (2. * pi * sd)
    w_sz = float(int(ns * st * sampling_rate))  # half time window size
    t = arange(-w_sz, w_sz + 1, dtype=float) / sampling_rate
    w = (exp(-t ** 2 / (2. * st ** 2)) * exp(2j * pi * f0 * t) /
         sqrt(sqrt(pi) * st * sampling_rate))
    return w


def _realwavelets(s_freq, freqs, dur, width):
    """Create real wavelets, for UCSD.

    Parameters
    ----------
    s_freq : int
        sampling frequency
    freqs : ndarray
        vector with frequencies of interest
    dur : float
        duration of the wavelets in s
    width : float
        parameter controlling gaussian shape

    Returns
    -------
    ndarray
        wavelets

    """
    x = arange(-dur/2, dur/2, 1/s_freq)
    wavelets = empty((len(freqs), len(x)))

    g = exp(-(pi * x ** 2) / width ** 2)

    for i, one_freq in enumerate(freqs):
        y = cos(2 * pi * x * one_freq)
        wavelets[i, :] = y * g

    return wavelets


def tukeywin(window_length, alpha=0.5):
    """Taken from http://leohart.wordpress.com/2006/01/29/hello-world/ """
    x = linspace(0, 1, window_length)
    w = ones(x.shape)

    # first condition 0 <= x < alpha/2
    first_condition = x < alpha / 2
    w[first_condition] = 0.5 * (1 + cos(2 * pi / alpha *
                                        (x[first_condition] - alpha / 2)))

    # second condition already taken care of

    # third condition 1 - alpha / 2 <= x <= 1
    third_condition = x >= (1 - alpha / 2)
    w[third_condition] = 0.5 * (1 + cos(2 * pi / alpha *
                                        (x[third_condition] - 1 + alpha / 2)))

    return w / len(w)



"""Module to plot all the elements as lines.
"""
from numpy import nanmax, nanmin
from vispy.scene.visuals import Rectangle

from .base import Viz


ONE_CHANNEL_HEIGHT = 30


class Viz1(Viz):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._limits_x = None  # tuple
        self._limits_y = None  # tuple

    def add_data(self, data, trial=0, axis_x='time', axis_subplot='chan',
                 limits_x=None, limits_y=None, color='k'):
        """
        Parameters
        ----------
        data : any instance of DataType
            Duck-typing should help
        trial : int
            index of the trial to plot
        axis_x : str, optional
            value to plot on x-axis, such as 'time' or 'freq'
        axis_subplot : str, optional
            axis to use for subplot
        limits_x : tuple, optional
            limits on the x-axis (if unspecified, it's the max across subplots)
        limits_y : tuple, optional
            limits on the y-axis (if unspecified, it's the max across subplots)
        """
        x = data.axis[axis_x][trial]
        max_x = nanmax(x)
        min_x = nanmin(x)

        subplot_values = data.axis[axis_subplot][trial]

        max_y = 0
        min_y = 0

        for cnt, one_value in enumerate(subplot_values):
            selected_axis = {axis_subplot: one_value}
            dat = data(trial=trial, **selected_axis)

            max_y = nanmax((max_y, nanmax(dat)))
            min_y = nanmin((min_y, nanmin(dat)))

            plt = self._fig[cnt, 0]
            plt.name = one_value
            plt.plot((x, dat), color=color, ylabel=one_value)

        if limits_x is not None:
            min_x, max_x = limits_x
        if limits_y is not None:
            min_y, max_y = limits_y
        for plt in self._fig.plot_widgets:
            plt.view.camera.set_range(x=(min_x, max_x), y=(min_y, max_y))

        self._limits_x = min_x, max_x
        self._limits_y = min_y, max_y

    def add_graphoelement(self, graphoelement, color='r', height=100):
        """Add graphoelements (at the moment, only spindles, but it works fine)

        Parameters
        ----------
        graphoelement : instance of Spindles
            the detected spindles
        color : str or 3-, 4- tuple
            color to use for the area of detection.
        height : float
            height of the highlight area

        Todo
        ----
            testing todo
        """
        for one_sp in graphoelement:
            chan = one_sp['chan']  # it could be over multiple channels
            start_time = one_sp['start_time']
            end_time = one_sp['end_time']

            if end_time > self._limits_x[0] and start_time < self._limits_x[1]:
                if start_time < self._limits_x[0]:
                    start_time = self._limits_x[0]
                if end_time > self._limits_x[1]:
                    end_time = self._limits_x[1]

                for plt in self._fig.plot_widgets:
                    if plt.name and plt.name in chan:
                        time_center = (end_time + start_time) / 2
                        width = end_time - time_center
                        rect = Rectangle(center=(time_center, 0), color=color,
                                         height=height, width=width)
                        rect.order = -1
                        plt.view.add(rect)
